from abc import ABC, abstractmethod
from typing import Any

from pydantic import BaseModel, Field


class InsertRowColumn(BaseModel):
    name: str = Field(..., title="Column name")
    value: str = Field(..., title="Column value")


class DeleteRowPk(BaseModel):
    name: str = Field(..., title="Primary key column name")
    value: str = Field(..., title="Primary key column value")


class BaseDataMutator(ABC):

    @abstractmethod
    def insert_row(self, table_name: str, row: list[InsertRowColumn]):
        """Insert a row in the table."""
        pass

    @abstractmethod
    def delete_row(self, table_name: str, pks: list[DeleteRowPk]):
        """Delete a row in the table. Row identified by primary key subset."""
        pass

    @abstractmethod
    def raw_query(self, query: str) -> list[Any]:
        """Execute a raw sql query in the database."""
        pass

    @abstractmethod
    def get_rows(self, table_name: str, limit: int = 100, offset: int = 0) -> list[Any]:
        """Get paginated rows from the table."""
        pass



from abc import ABC
import enum
from dataclasses import dataclass, field


class ConstraintType(enum.Enum):
    """All supported constraint types"""

    NOT_NULL = enum.auto()
    UNIQUE = enum.auto()
    PRIMARY_KEY = enum.auto()
    FOREIGN_KEY = enum.auto()
    CHECK = enum.auto()
    DEFAULT = enum.auto()


@dataclass(frozen=True)
class Constraint(ABC):
    """Base class for all constraints"""

    constraint_type: ConstraintType = field(init=False)
    table_name: str


@dataclass(frozen=True)
class NotNullConstraint(Constraint):
    """Class for the NOT NULL constraint"""

    constraint_type: ConstraintType = field(init=False, default=ConstraintType.NOT_NULL)
    column_name: str


@dataclass(frozen=True)
class UniqueConstraint(Constraint):
    """Class for the UNIQUE constraint"""

    constraint_type: ConstraintType = field(init=False, default=ConstraintType.UNIQUE)
    column_name: str


@dataclass(frozen=True)
class PrimaryKeyConstraint(Constraint):
    """Class for the PRIMARY KEY constraint"""

    constraint_type: ConstraintType = field(
        init=False, default=ConstraintType.PRIMARY_KEY
    )
    column_name: str


@dataclass(frozen=True, unsafe_hash=True)
class ForeignKeyConstraint(Constraint):
    """Class for the FOREIGN KEY constraint"""

    constraint_type: ConstraintType = field(
        init=False, default=ConstraintType.FOREIGN_KEY
    )
    referenced_table: str
    column_mapping: dict[str, str] = field(
        hash=False
    )  # maps the corresponding column names: {referencing_column1: referenced_column1, ...}


@dataclass(frozen=True)
class CheckConstraint(Constraint):
    """Class for the CHECK constraint"""

    constraint_type: ConstraintType = field(init=False, default=ConstraintType.CHECK)
    column_name: str
    condition: str


@dataclass(frozen=True)
class DefaultConstraint(Constraint):
    """Class for the DEFAULT constraint"""

    constraint_type: ConstraintType = field(init=False, default=ConstraintType.DEFAULT)
    column_name: str
    value: str



from skibidi_orm.migration_engine.data_mutator.base_data_mutator import BaseDataMutator
from skibidi_orm.migration_engine.data_mutator.sqlite3_data_mutatorr import (
    SQLite3DataMutator,
)
from skibidi_orm.migration_engine.db_inspectors.base_inspector import BaseDbInspector
from skibidi_orm.migration_engine.db_inspectors.sqlite3_inspector import (
    SQLite3Inspector,
)


def get_db_mutator(db_inspector: BaseDbInspector) -> BaseDataMutator:
    if isinstance(db_inspector, SQLite3Inspector):
        return SQLite3DataMutator()
    raise NotImplementedError



import sqlite3
from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
from skibidi_orm.migration_engine.sql_executor.base_sql_executor import BaseSQLExecutor
from skibidi_orm.migration_engine.operations.column_operations import ColumnOperation
from skibidi_orm.migration_engine.operations.table_operations import TableOperation

from skibidi_orm.migration_engine.converters.sqlite3.all import SQLite3Converter


class SQLite3Executor(BaseSQLExecutor):

    @staticmethod
    def execute_sql(sql: str):
        sqlite_config = SQLite3Config.get_instance()
        with sqlite3.connect(sqlite_config.db_path) as conn:
            cursor = conn.cursor()
            commands = sql.split(";")
            for command in commands:
                if not command:
                    continue
                cursor.execute(command.strip())
            conn.commit()

    @staticmethod
    def execute_operations(operations: list[TableOperation | ColumnOperation]):
        for operation in operations:
            if isinstance(operation, TableOperation):
                SQLite3Executor.execute_sql(
                    SQLite3Converter.convert_table_operation_to_SQL(operation)
                )
            else:
                SQLite3Executor.execute_sql(
                    SQLite3Converter.convert_column_operation_to_SQL(operation)
                )



from abc import ABC, abstractmethod
from skibidi_orm.migration_engine.operations.table_operations import TableOperation
from skibidi_orm.migration_engine.operations.column_operations import ColumnOperation


class IStateManager(ABC):

    @abstractmethod
    def get_operations(self) -> list[TableOperation | ColumnOperation]:
        """Get operations needed to transform database structure into schema structure"""



# Example of how Table class and classes modeling DB schema can be implemented
from skibidi_orm.migration_engine.adapters.base_adapter import BaseAdapter
from skibidi_orm.migration_engine.adapters.database_objects.sqlite3_typing import (
    SQLite3Typing,
)
from skibidi_orm.migration_engine.adapters.sqlite3_adapter import SQLite3Adapter
import skibidi_orm.migration_engine.adapters.database_objects.constraints as c
from skibidi_orm.migration_engine.adapters.database_objects.migration_element import (
    MigrationElement,
)


class Table(MigrationElement):
    """
    This is example of how Table class can be implemented
    Key notes, that have to be taken into account:
    - Table class HAVE to inherit from MigrationElement and MigrationElement MUST only have one class inheriting from it
    - Table class HAVE to have adapter attribute, which is instance of BaseAdapter or its child
    """

    adapter: BaseAdapter

    def __init__(self) -> None:
        self.adapter = SQLite3Adapter()

        # This will iterate over all classes that inherit from Table and create a table in the database
        models = Table.__subclasses__()

        # Creator of Table class can implement how to map their own class structure to fit the adapter

        # This will only populate the adapter in inside migration scripts, as Table will never be instantiated
        if self.__class__ == Table:
            for cls in models:
                table: SQLite3Typing.Table
                data_type: SQLite3Typing.DataTypes
                columns: list[SQLite3Typing.Column] = []
                all_constraints = cls.__dict__["constraints"]

                fks: set[c.ForeignKeyConstraint] = set(
                    constraint
                    for constraint in all_constraints
                    if isinstance(constraint, c.ForeignKeyConstraint)
                )

                column_constraints = [
                    constraint
                    for constraint in all_constraints
                    if isinstance(constraint, c.ColumnSpecificConstraint)
                ]

                if cls.__dict__["data_type"] == "my_definition_of_data_type":
                    data_type = "INTEGER"
                elif cls.__dict__["data_type"] == "my_other_definition":
                    data_type = "TEXT"
                elif cls.__dict__["data_type"] == "and_other":
                    data_type = "BLOB"
                else:
                    data_type = "NULL"

                columns.append(
                    SQLite3Typing.Column(
                        name=cls.__dict__["name"],
                        data_type=data_type,
                        column_constraints=column_constraints,
                    )
                )
                table = SQLite3Typing.Table(
                    name=cls.__name__, columns=columns, foreign_keys=fks
                )

                # Table, for each instance should properly inform the adapter about being created
                self.adapter.create_table(table)


class User(Table):
    name = "test_name"
    data_type = "my_definition_of_data_type"
    constraints = None
    data: list[tuple[str, str]] = []

    def insert_data(self, name: str, title: str):
        self.data.append((name, title))


class Post(Table):
    name = "test_title"
    data_type = "my_other_definition"
    constraints = None


class Comment(Table):
    name = "test_content"
    data_type = "and_other"
    constraints = None



from abc import ABC
import enum
from dataclasses import dataclass, field
from functools import total_ordering
from typing import Any


class ConstraintType(enum.Enum):
    """All supported constraint types"""

    NOT_NULL = "NOT NULL"
    UNIQUE = "UNIQUE"
    PRIMARY_KEY = "PRIMARY KEY"
    FOREIGN_KEY = "FOREIGN KEY"
    CHECK = "CHECK"
    DEFAULT = "DEFAULT"


@dataclass(frozen=True)
class Constraint(ABC):
    """Base class for all constraints"""

    constraint_type: ConstraintType = field(init=False)
    table_name: str


@total_ordering
@dataclass(frozen=True)
class ColumnSpecificConstraint(Constraint):
    """Base class for constraints that apply to a column instead of multiple (e.g. composite foreign keys)"""

    column_name: str

    def __lt__(self, other: Any):
        if isinstance(other, ColumnSpecificConstraint):
            return (self.column_name + self.__class__.__name__) < (
                other.column_name + other.__class__.__name__
            )
        return NotImplemented


@dataclass(frozen=True)
class NotNullConstraint(ColumnSpecificConstraint):
    """Class for the NOT NULL constraint"""

    constraint_type: ConstraintType = field(init=False, default=ConstraintType.NOT_NULL)


@dataclass(frozen=True)
class UniqueConstraint(ColumnSpecificConstraint):
    """Class for the UNIQUE constraint"""

    constraint_type: ConstraintType = field(init=False, default=ConstraintType.UNIQUE)


@dataclass(frozen=True)
class PrimaryKeyConstraint(ColumnSpecificConstraint):
    """Class for the PRIMARY KEY constraint"""

    constraint_type: ConstraintType = field(
        init=False, default=ConstraintType.PRIMARY_KEY
    )


@dataclass(frozen=True, unsafe_hash=True)
class ForeignKeyConstraint(Constraint):
    """Class for the FOREIGN KEY constraint"""

    constraint_type: ConstraintType = field(
        init=False, default=ConstraintType.FOREIGN_KEY
    )
    referenced_table: str
    column_mapping: dict[str, str] = field(
        hash=False
    )  # maps the corresponding column names: {referencing_column1: referenced_column1, ...}


@dataclass(frozen=True)
class CheckConstraint(ColumnSpecificConstraint):
    """Class for the CHECK constraint"""

    constraint_type: ConstraintType = field(init=False, default=ConstraintType.CHECK)
    condition: str


@dataclass(frozen=True)
class DefaultConstraint(ColumnSpecificConstraint):
    """Class for the DEFAULT constraint"""

    constraint_type: ConstraintType = field(init=False, default=ConstraintType.DEFAULT)
    value: str



import importlib
from os import path
import sys
from skibidi_orm.migration_engine.db_config.base_config import BaseDbConfig


def db_config_dynamic_import(schema_file_path: str) -> BaseDbConfig:
    if not path.exists(schema_file_path):
        raise FileNotFoundError(f"Schema file {schema_file_path} not found")
    if not path.isfile(schema_file_path):
        raise TypeError(f"Schema file {schema_file_path} is not a file")
    file = path.basename(schema_file_path)
    dir = path.dirname(schema_file_path)
    sys.path.append(dir)
    importlib.import_module(file[:-3])

    return BaseDbConfig.get_instance()



import skibidi_orm.migration_engine.adapters.database_objects.constraints as c


def test_ColumnSpecificConstraint_sortable():
    sorted(
        [
            c.PrimaryKeyConstraint(
                table_name="table_name",
                column_name="column_name",
            ),
            c.NotNullConstraint(
                table_name="table_name",
                column_name="column_name",
            ),
            c.CheckConstraint(
                table_name="table_name",
                column_name="column_name",
                condition="condition",
            ),
        ]
    )
    assert True



from __future__ import annotations

from typing import Any, cast

import skibidi_orm.migration_engine.adapters.database_objects.constraints as c
from skibidi_orm.migration_engine.adapters.postgres_typing import PostgresTyping
from skibidi_orm.migration_engine.db_config.postgres_config import PostgresConfig
from skibidi_orm.migration_engine.db_inspectors.base_inspector import BaseDbInspector


class PostgresInspector(BaseDbInspector):
    """
    Used to get data from live Postgres database.
    Should only be instantiated when Postgres is choosen as the database.
    """

    def __init__(self) -> None:
        self.config = PostgresConfig.get_instance()

    def get_tables_names(self) -> list[str]:
        """
        Get all tables names from the database.
        """

        with self.config.connection.cursor() as cursor:
            cursor.execute(
                """
                SELECT table_name
                FROM information_schema.tables
                WHERE table_schema = 'public'
                """
            )

            tables = cursor.fetchall()

        return [table[0] for table in tables]

    def get_tables(self) -> list[PostgresTyping.Table]:
        """
        Get all tables from the database.
        """

        return [
            PostgresTyping.Table(
                name=table_name,
                columns=self.get_table_columns(table_name),
                foreign_keys=self._get_foreign_keys(table_name),
            )
            for table_name in self.get_tables_names()
        ]

    def get_table_columns(self, table_name: str) -> list[PostgresTyping.Column]:
        """
        Get all columns from the table.
        """

        return [
            PostgresTyping.Column(
                name=column_name,
                data_type=self._get_column_data_type(table_name, column_name),
                column_constraints=self._get_column_constraints(
                    table_name, column_name
                ),
            )
            for column_name in self._get_table_columns_names(table_name)
        ]

    def _get_foreign_keys(self, table_name: str) -> set[c.ForeignKeyConstraint]:
        """
        Get all foreign keys from the table.
        """

        with self.config.connection.cursor() as cursor:
            cursor.execute(
                f"""
                SELECT
                    tc.constraint_name,
                    tc.table_name,
                    kcu.column_name,
                    ccu.table_name AS foreign_table_name,
                    ccu.column_name AS foreign_column_name
                FROM
                    information_schema.table_constraints AS tc
                JOIN
                    information_schema.key_column_usage AS kcu
                    ON tc.constraint_name = kcu.constraint_name
                JOIN
                    information_schema.constraint_column_usage AS ccu
                    ON tc.constraint_name = ccu.constraint_name
                WHERE
                    tc.constraint_type = 'FOREIGN KEY'
                    AND tc.table_name = '{table_name}';
                """
            )

            rows = cursor.fetchall()

        res: set[c.ForeignKeyConstraint] = set()
        for row in rows:
            (
                _,
                table_name,
                column_name,
                foreign_table_name,
                foreign_column_name,
            ) = row

            res.add(
                c.ForeignKeyConstraint(
                    table_name=table_name,
                    referenced_table=foreign_table_name,
                    column_mapping={column_name: foreign_column_name},
                )
            )

        return res

    def _get_column_constraints(
        self, table_name: str, column_name: str
    ) -> list[c.ColumnSpecificConstraint]:
        """
        Get all column constraints from the table.
        """

        with self.config.connection.cursor() as cursor:
            cursor.execute(
                f"""
                SELECT
                    col.table_name,
                    col.column_name,
                    tc.constraint_type,
                    tc.constraint_name,
                    cc.check_clause,
                    col.column_default
                FROM
                    information_schema.columns AS col
                LEFT JOIN
                    information_schema.constraint_column_usage AS ccu
                    ON col.table_schema = ccu.table_schema
                    AND col.table_name = ccu.table_name
                    AND col.column_name = ccu.column_name
                LEFT JOIN
                    information_schema.table_constraints AS tc
                    ON ccu.table_schema = tc.table_schema
                    AND ccu.table_name = tc.table_name
                    AND ccu.constraint_name = tc.constraint_name
                LEFT JOIN
                    information_schema.check_constraints AS cc
                    ON tc.constraint_name = cc.constraint_name
                WHERE
                    col.table_name = '{table_name}'
                    AND col.column_name = '{column_name}'
                ORDER BY
                    col.table_name,
                    col.ordinal_position;
                """
            )

            rows = cursor.fetchall()
            if len(rows) == 0:
                raise ValueError(
                    f"Column '{column_name}' does not exist in table '{table_name}'"
                )

        res: list[c.ColumnSpecificConstraint] = []
        for row in rows:  # loop because we can have many constraints for one column
            _, _, constraint_type, _, check_clause, column_default = row

            if check_clause:
                res.append(c.CheckConstraint(table_name, column_name, check_clause))
            if constraint_type == "PRIMARY KEY":
                res.append(c.PrimaryKeyConstraint(table_name, column_name))
            if constraint_type == "UNIQUE":
                res.append(c.UniqueConstraint(table_name, column_name))
            if not self._is_column_nullable(
                table_name, column_name
            ) and not self.__holds_instance(res, c.NotNullConstraint):
                res.append(c.NotNullConstraint(table_name, column_name))
            if column_default and not self.__holds_instance(res, c.DefaultConstraint):
                res.append(c.DefaultConstraint(table_name, column_name, column_default))

        return res

    def _get_table_columns_names(self, table_name: str) -> list[str]:
        """
        Get all columns names from the table.
        """

        with self.config.connection.cursor() as cursor:
            cursor.execute(
                f"""
                SELECT column_name
                FROM information_schema.columns
                WHERE table_name = '{table_name}'
                """
            )

            columns = cursor.fetchall()

        return [column[0] for column in columns]

    def _get_column_data_type(
        self, table_name: str, column_name: str
    ) -> PostgresTyping.DataTypes:
        """
        Get the data type of the column.
        """
        with self.config.connection.cursor() as cursor:
            cursor.execute(
                f"""
                SELECT data_type
                FROM information_schema.columns
                WHERE table_name = '{table_name}'
                AND column_name = '{column_name}'
                """
            )

            data_type = cursor.fetchone()
            if not data_type:
                raise ValueError(
                    f"Column '{column_name}' does not exist in table '{table_name}'"
                )

        data_type_upper = cast(str, data_type[0]).upper()
        return cast(PostgresTyping.DataTypes, data_type_upper)

    def _is_column_nullable(self, table_name: str, column_name: str) -> bool:
        """
        Check if the column is nullable.
        """

        with self.config.connection.cursor() as cursor:
            cursor.execute(
                f"""
                SELECT is_nullable
                FROM information_schema.columns
                WHERE table_name = '{table_name}'
                AND column_name = '{column_name}'
                """
            )

            is_nullable = cursor.fetchone()
            if not is_nullable:
                raise ValueError(
                    f"Column '{column_name}' does not exist in table '{table_name}'"
                )

        return is_nullable[0] == "YES"

    def __holds_instance(self, array: list[Any], class_type: Any) -> bool:
        """
        Helper function to check if the array holds the instance of a class.
        """
        return any(isinstance(x, class_type) for x in array)



from typing import Literal
from skibidi_orm.migration_engine.adapters.base_adapter import (
    BaseColumn,
    BaseTable,
)

from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    Constraint,
)


class SQLite3Typing:

    DataTypes = Literal["TEXT", "INTEGER", "REAL", "BLOB", "NULL"]
    Constraints = Constraint
    Column = BaseColumn[DataTypes]
    Table = BaseTable[Column]
    tables: list[Table] = []



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from functools import total_ordering
from typing import Any

from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    ColumnSpecificConstraint,
    ForeignKeyConstraint,
)


@total_ordering
@dataclass(unsafe_hash=True)
class BaseColumn[TDataTypes]:
    name: str
    data_type: TDataTypes
    column_constraints: list[ColumnSpecificConstraint] = field(
        default_factory=list, hash=False
    )

    def __lt__(self, other: Any):
        if isinstance(other, BaseColumn):
            return self.name < other.name
        return NotImplemented


@dataclass
class BaseTable[TCol]:
    name: str
    columns: list[TCol] = field(default_factory=list)
    foreign_keys: set[ForeignKeyConstraint] = field(default_factory=set)


class BaseAdapter(ABC):

    @abstractmethod
    def create_table(self, table: BaseTable[BaseColumn[Any]]):
        """Create a table in the database"""
        pass

    @abstractmethod
    def execute_migration(self, preview: bool = False):
        """Execute the migration"""
        pass



from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
from skibidi_orm.migration_engine.db_inspectors.sqlite3_inspector import SQLite3Inspector
from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    PrimaryKeyConstraint,
)
from skibidi_orm.migration_engine.sql_executor.sqlite3_executor import SQLite3Executor


def test_execute_sql(make_database: str):
    SQLite3Config(make_database)
    SQLite3Executor.execute_sql(
        "CREATE TABLE test_table (id INTEGER PRIMARY KEY, name TEXT);"
    )
    table = SQLite3Inspector().get_tables()[0]
    assert table.name == "test_table"
    assert table.columns[0].name == "id"
    assert table.columns[0].data_type == "INTEGER"
    assert table.columns[0].column_constraints == [PrimaryKeyConstraint("test_table", "id")]
    assert table.columns[1].name == "name"
    assert table.columns[1].data_type == "TEXT"
    assert table.columns[1].column_constraints == []



import pathlib
import pytest
from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
from skibidi_orm.migration_engine.adapters.database_objects import constraints as c
import sqlite3

from skibidi_orm.migration_engine.db_inspectors.sqlite3_inspector import (
    SQLite3Inspector,
)

from ..sql_data import SQLite3TablesData


def execute_sqlite3_commands(db_path: str, commands: list[str]):
    """Executes the given commands in a SQLite DB living under db_path"""
    with sqlite3.connect(db_path) as conn:
        cursor = conn.cursor()
        for command in commands:
            cursor.execute(command)
        conn.commit()
        cursor.close()


@pytest.fixture
def tmp_database(request: pytest.FixtureRequest, tmp_path: pathlib.Path):
    """Fixture for creating a temporary database and executing sqlite3 code. Yields the
    path of the db file."""
    sql_commands = request.param
    (tmp_file := tmp_path.joinpath("tmp_db.db")).touch()
    execute_sqlite3_commands(
        str(tmp_file),
        sql_commands,
    )
    yield str(tmp_file)


@pytest.mark.parametrize(
    "tmp_database",
    [[SQLite3TablesData.sql_table1, SQLite3TablesData.sql_table2]],
    indirect=True,
)
@pytest.mark.usefixtures("tmp_database")
def test_can_only_be_instantiated_with_sqlite3config_instantiated_earlier():
    with pytest.raises(ReferenceError) as exc_info:
        SQLite3Inspector()
    assert str(exc_info.value) == "Instance does not exist"


@pytest.mark.parametrize(
    "tmp_database",
    [[SQLite3TablesData.sql_table1, SQLite3TablesData.sql_table2]],
    indirect=True,
)
def test_get_tables_names(tmp_database: str):
    SQLite3Config(tmp_database)
    inspector = SQLite3Inspector()
    tables = inspector.get_tables_names()
    assert len(tables) == 2
    assert tables[0] == "table1"
    assert tables[1] == "table2"


@pytest.mark.parametrize(
    "tmp_database",
    [[SQLite3TablesData.sql_table1, SQLite3TablesData.sql_table2]],
    indirect=True,
)
def test_get_table_columns(tmp_database: str):
    SQLite3Config(db_path=tmp_database)
    inspector = SQLite3Inspector()
    columns = inspector.get_table_columns("table1")
    assert columns[0].name == "id"
    assert columns[0].data_type == "INTEGER"
    assert columns[0].column_constraints == [c.PrimaryKeyConstraint("table1", "id")]
    assert columns[1].name == "name"
    assert columns[1].data_type == "TEXT"
    assert columns[1].column_constraints == [c.NotNullConstraint("table1", "name")]
    assert len(columns) == 2


@pytest.mark.parametrize(
    "tmp_database", [[SQLite3TablesData.sql_table_primary_key_not_null]], indirect=True
)
def test_get_table_columns__primaryk_notnull(tmp_database: str):
    SQLite3Config(db_path=tmp_database)
    inspector = SQLite3Inspector()
    columns = inspector.get_table_columns("table_primary_key_not_null")
    assert columns[0].name == "id"
    assert columns[0].data_type == "INTEGER"
    assert columns[0].column_constraints == [
        c.PrimaryKeyConstraint("table_primary_key_not_null", "id"),
        c.NotNullConstraint("table_primary_key_not_null", "id"),
    ]


@pytest.mark.parametrize(
    "tmp_database",
    [[SQLite3TablesData.sql_table1, SQLite3TablesData.sql_table2]],
    indirect=True,
)
def test_get_tables(tmp_database: str):
    SQLite3Config(db_path=tmp_database)
    inspector = SQLite3Inspector()
    tables = inspector.get_tables()
    assert len(tables) == 2
    assert tables[0].name == "table1"
    assert tables[1].name == "table2"
    assert len(tables[0].columns) == 2
    assert len(tables[1].columns) == 2
    assert tables[0].columns[0].name == "id"
    assert tables[0].columns[0].data_type == "INTEGER"
    assert tables[0].columns[0].column_constraints == [
        c.PrimaryKeyConstraint("table1", "id")
    ]
    assert tables[0].columns[1].name == "name"
    assert tables[0].columns[1].data_type == "TEXT"
    assert tables[0].columns[1].column_constraints == [
        c.NotNullConstraint("table1", "name")
    ]
    assert tables[1].columns[0].name == "id"
    assert tables[1].columns[0].data_type == "INTEGER"
    assert tables[1].columns[0].column_constraints == [
        c.PrimaryKeyConstraint("table2", "id")
    ]
    assert tables[1].columns[1].name == "name"
    assert tables[1].columns[1].data_type == "TEXT"
    assert tables[1].columns[1].column_constraints == [
        c.NotNullConstraint("table2", "name")
    ]


@pytest.mark.parametrize(
    "tmp_database", [[*SQLite3TablesData.sql_schema_with_fks]], indirect=True
)
def test_get_foreign_keys(tmp_database: str):
    SQLite3Config(db_path=tmp_database)
    inspector = SQLite3Inspector()
    foreign_keys = inspector.get_foreign_key_constraints()
    assert len(foreign_keys) == 3
    correct_fk_set = {
        c.ForeignKeyConstraint("posts", "users", {"user_id": "user_id"}),
        c.ForeignKeyConstraint(
            "comments", "users", {"user_idd": "user_id", "username": "username"}
        ),
        c.ForeignKeyConstraint("comments", "posts", {"post_id": "post_id"}),
    }
    assert correct_fk_set.intersection(foreign_keys) == foreign_keys



import glob
from skibidi_orm.exceptions.cli_exceptions import MultipleSchemaFilesError

GLOB_ALL_SCHEMAS = "**/**/schema.py"


def find_schema_file():
    """
    Find the schema file in the project.
    If no schema file is found, raise FileNotFoundError.
    If multiple schema files are found, raise MultipleSchemaFilesError.
    """
    files = set(glob.glob(GLOB_ALL_SCHEMAS, recursive=True))
    if len(files) == 0:
        raise FileNotFoundError("No schema file found.")
    if len(files) > 1:
        raise MultipleSchemaFilesError("Multiple schema files found.")
    return files.pop()



from skibidi_orm.migration_engine.db_config.postgres_config import PostgresConfig
from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
from skibidi_orm.migration_engine.db_inspectors.base_inspector import BaseDbInspector
from skibidi_orm.migration_engine.db_inspectors.sqlite3_inspector import (
    SQLite3Inspector,
)
from skibidi_orm.migration_engine.studio.utils.db_config_dynamic_import import (
    db_config_dynamic_import,
)


def get_db_inspector(schema_file: str) -> BaseDbInspector:
    db_config = db_config_dynamic_import(schema_file_path=schema_file)
    if isinstance(db_config, SQLite3Config):
        return SQLite3Inspector()
    elif isinstance(db_config, PostgresConfig):
        raise NotImplementedError

    raise NotImplementedError



from skibidi_orm.migration_engine.adapters.database_objects.sqlite3_typing import (
    SQLite3Typing,
)
from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
from skibidi_orm.migration_engine.db_inspectors.sqlite3_inspector import (
    SQLite3Inspector,
)
from skibidi_orm.migration_engine.sql_executor.sqlite3_executor import SQLite3Executor
from skibidi_orm.migration_engine.adapters.sqlite3_adapter import SQLite3Adapter
from skibidi_orm.migration_engine.adapters.database_objects.migration_element import (
    MigrationElement,
)
from skibidi_orm.migration_engine.adapters.database_objects import constraints as c
from .sql_data import SQLite3TablesData


def test_adding_table_to_database(make_database: str):
    SQLite3Config(make_database)
    SQLite3Executor.execute_sql(SQLite3TablesData.sql_table_user)
    SQLite3Executor.execute_sql(SQLite3TablesData.sql_table_post)

    class Table(MigrationElement):
        def __init__(self) -> None:
            self.adapter = SQLite3Adapter()

            models = Table.__subclasses__()
            if self.__class__ == Table:
                for cls in models:
                    self.adapter.create_table(cls.__dict__["table"])

    class User(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="user_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("User", "user_id")],
            ),
            SQLite3Typing.Column(
                name="user_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("User", "user_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="User", columns=columns)

    class Post(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="post_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("Post", "post_id")],
            ),
            SQLite3Typing.Column(
                name="post_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("Post", "post_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="Post", columns=columns)

    class Comment(Table):  # type: ignore

        columns = [
            SQLite3Typing.Column(
                name="comment_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("Comment", "comment_id")],
            ),
            SQLite3Typing.Column(
                name="comment_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("Comment", "comment_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="Comment", columns=columns)

    m = MigrationElement()
    m.migrate()

    tables = SQLite3Inspector().get_tables()

    assert len(tables) == 3
    assert tables[2].name == "Comment"
    assert tables[2].columns[0].name == "comment_id"
    assert tables[2].columns[0].data_type == "INTEGER"
    assert tables[2].columns[0].column_constraints == [
        c.PrimaryKeyConstraint("Comment", "comment_id")
    ]
    assert tables[2].columns[1].name == "comment_name"
    assert tables[2].columns[1].data_type == "TEXT"
    assert tables[2].columns[1].column_constraints == [
        c.NotNullConstraint("Comment", "comment_name")
    ]


def test_removing_table_from_database(make_database: str):
    SQLite3Config(make_database)
    SQLite3Executor.execute_sql(SQLite3TablesData.sql_table_user)
    SQLite3Executor.execute_sql(SQLite3TablesData.sql_table_post)
    SQLite3Executor.execute_sql(SQLite3TablesData.sql_table_comment)

    class Table(MigrationElement):

        def __init__(self) -> None:
            self.adapter = SQLite3Adapter()

            models = Table.__subclasses__()
            if self.__class__ == Table:
                for cls in models:
                    self.adapter.create_table(cls.__dict__["table"])

    class User(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="user_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("User", "user_id")],
            ),
            SQLite3Typing.Column(
                name="user_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("User", "user_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="User", columns=columns)

    class Post(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="post_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("Post", "post_id")],
            ),
            SQLite3Typing.Column(
                name="post_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("Post", "post_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="Post", columns=columns)

    m = MigrationElement()
    m.migrate()

    tables = SQLite3Inspector().get_tables()

    tables.sort(key=lambda table: table.name, reverse=True)

    assert len(tables) == 2
    assert tables[0].name == "User"
    assert tables[0].columns[0].name == "user_id"
    assert tables[0].columns[0].data_type == "INTEGER"
    assert tables[0].columns[0].column_constraints == [
        c.PrimaryKeyConstraint("User", "user_id")
    ]
    assert tables[0].columns[1].name == "user_name"
    assert tables[0].columns[1].data_type == "TEXT"
    assert tables[0].columns[1].column_constraints == [
        c.NotNullConstraint("User", "user_name")
    ]

    assert tables[1].name == "Post"
    assert tables[1].columns[0].name == "post_id"
    assert tables[1].columns[0].data_type == "INTEGER"
    assert tables[1].columns[0].column_constraints == [
        c.PrimaryKeyConstraint("Post", "post_id")
    ]
    assert tables[1].columns[1].name == "post_name"
    assert tables[1].columns[1].data_type == "TEXT"
    assert tables[1].columns[1].column_constraints == [
        c.NotNullConstraint("Post", "post_name")
    ]


def test_add_column_to_database(make_database: str):
    SQLite3Config(make_database)
    SQLite3Executor.execute_sql(SQLite3TablesData.sql_table_user)
    SQLite3Executor.execute_sql(SQLite3TablesData.sql_table_post)

    class Table(MigrationElement):

        def __init__(self) -> None:
            self.adapter = SQLite3Adapter()

            models = Table.__subclasses__()
            if self.__class__ == Table:
                for cls in models:
                    self.adapter.create_table(cls.__dict__["table"])

    class User(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="user_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("User", "user_id")],
            ),
            SQLite3Typing.Column(
                name="user_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("User", "user_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="User", columns=columns)

    class Post(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="post_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("Post", "post_id")],
            ),
            SQLite3Typing.Column(
                name="post_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("Post", "post_name")],
            ),
            SQLite3Typing.Column(
                name="post_content",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("Post", "post_content")],
            ),
        ]

        table = SQLite3Typing.Table(name="Post", columns=columns)

    m = MigrationElement()
    m.migrate()

    tables = SQLite3Inspector().get_tables()

    tables.sort(key=lambda table: table.name, reverse=True)

    assert len(tables) == 2
    assert tables[0].name == "User"
    assert tables[0].columns[0].name == "user_id"
    assert tables[0].columns[0].data_type == "INTEGER"
    assert tables[0].columns[0].column_constraints == [
        c.PrimaryKeyConstraint("User", "user_id")
    ]
    assert tables[0].columns[1].name == "user_name"
    assert tables[0].columns[1].data_type == "TEXT"
    assert tables[0].columns[1].column_constraints == [
        c.NotNullConstraint("User", "user_name")
    ]

    assert tables[1].name == "Post"
    assert tables[1].columns[0].name == "post_id"
    assert tables[1].columns[0].data_type == "INTEGER"
    assert tables[1].columns[0].column_constraints == [
        c.PrimaryKeyConstraint("Post", "post_id")
    ]
    assert tables[1].columns[1].name == "post_name"
    assert tables[1].columns[1].data_type == "TEXT"
    assert tables[1].columns[1].column_constraints == [
        c.NotNullConstraint("Post", "post_name")
    ]
    assert tables[1].columns[2].name == "post_content"
    assert tables[1].columns[2].data_type == "TEXT"
    assert tables[1].columns[2].column_constraints == [
        c.NotNullConstraint("Post", "post_content")
    ]


def test_removing_column_from_database(make_database: str):
    SQLite3Config(make_database)
    SQLite3Executor.execute_sql(SQLite3TablesData.sql_table_user)
    SQLite3Executor.execute_sql(SQLite3TablesData.sql_table_post)

    class Table(MigrationElement):
        def __init__(self) -> None:
            self.adapter = SQLite3Adapter()

            models = Table.__subclasses__()
            if self.__class__ == Table:
                for cls in models:
                    self.adapter.create_table(cls.__dict__["table"])

    class User(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="user_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("User", "user_id")],
            ),
            SQLite3Typing.Column(
                name="user_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("User", "user_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="User", columns=columns)

    class Post(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="post_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("Post", "post_id")],
            ),
        ]

        table = SQLite3Typing.Table(name="Post", columns=columns)

    m = MigrationElement()
    m.migrate()

    tables = SQLite3Inspector().get_tables()

    tables.sort(key=lambda table: table.name, reverse=True)

    assert len(tables) == 2
    assert tables[0].name == "User"
    assert tables[0].columns[0].name == "user_id"
    assert tables[0].columns[0].data_type == "INTEGER"
    assert tables[0].columns[0].column_constraints == [
        c.PrimaryKeyConstraint("User", "user_id")
    ]
    assert tables[0].columns[1].name == "user_name"
    assert tables[0].columns[1].data_type == "TEXT"
    assert tables[0].columns[1].column_constraints == [
        c.NotNullConstraint("User", "user_name")
    ]

    assert tables[1].name == "Post"
    assert len(tables[1].columns) == 1
    assert tables[1].columns[0].name == "post_id"
    assert tables[1].columns[0].data_type == "INTEGER"
    assert tables[1].columns[0].column_constraints == [
        c.PrimaryKeyConstraint("Post", "post_id")
    ]


def test_renaming_table_in_database(make_database: str):
    SQLite3Config(make_database)
    SQLite3Executor.execute_sql(SQLite3TablesData.sql_table_user)
    SQLite3Executor.execute_sql(SQLite3TablesData.sql_table_post)

    class Table(MigrationElement):

        def __init__(self) -> None:
            self.adapter = SQLite3Adapter()

            models = Table.__subclasses__()
            if self.__class__ == Table:
                for cls in models:
                    self.adapter.create_table(cls.__dict__["table"])

    class User(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="user_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("User", "user_id")],
            ),
            SQLite3Typing.Column(
                name="user_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("User", "user_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="User", columns=columns)

    class NewNamePost(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="post_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("NewNamePost", "post_id")],
            ),
            SQLite3Typing.Column(
                name="post_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("NewNamePost", "post_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="NewNamePost", columns=columns)

    m = MigrationElement()
    m.migrate()

    tables = SQLite3Inspector().get_tables()

    assert len(tables) == 2
    assert tables[0].name == "User"
    assert tables[0].columns[0].name == "user_id"
    assert tables[0].columns[0].data_type == "INTEGER"
    assert tables[0].columns[0].column_constraints == [
        c.PrimaryKeyConstraint("User", "user_id")
    ]
    assert tables[0].columns[1].name == "user_name"
    assert tables[0].columns[1].data_type == "TEXT"
    assert tables[0].columns[1].column_constraints == [
        c.NotNullConstraint("User", "user_name")
    ]

    assert tables[1].name == "NewNamePost"
    assert tables[1].columns[0].name == "post_id"
    assert tables[1].columns[0].data_type == "INTEGER"
    assert tables[1].columns[0].column_constraints == [
        c.PrimaryKeyConstraint("NewNamePost", "post_id")
    ]
    assert tables[1].columns[1].name == "post_name"
    assert tables[1].columns[1].data_type == "TEXT"
    assert tables[1].columns[1].column_constraints == [
        c.NotNullConstraint("NewNamePost", "post_name")
    ]


def test_add_column_to_database_and_remove_table_from_database(make_database: str):
    SQLite3Config(make_database)
    SQLite3Executor.execute_sql(SQLite3TablesData.sql_table_user)
    SQLite3Executor.execute_sql(SQLite3TablesData.sql_table_post)
    SQLite3Executor.execute_sql(SQLite3TablesData.sql_table_comment)

    class Table(MigrationElement):

        def __init__(self) -> None:
            self.adapter = SQLite3Adapter()

            models = Table.__subclasses__()
            if self.__class__ == Table:
                for cls in models:
                    self.adapter.create_table(cls.__dict__["table"])

    class User(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="user_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("User", "user_id")],
            ),
            SQLite3Typing.Column(
                name="user_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("User", "user_name")],
            ),
        ]

        table = SQLite3Typing.Table(name="User", columns=columns)

    class Post(Table):  # type: ignore
        columns = [
            SQLite3Typing.Column(
                name="post_id",
                data_type="INTEGER",
                column_constraints=[c.PrimaryKeyConstraint("Post", "post_id")],
            ),
            SQLite3Typing.Column(
                name="post_name",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("Post", "post_name")],
            ),
            SQLite3Typing.Column(
                name="post_content",
                data_type="TEXT",
                column_constraints=[c.NotNullConstraint("Post", "post_content")],
            ),
        ]

        table = SQLite3Typing.Table(name="Post", columns=columns)

    m = MigrationElement()
    m.migrate()

    tables = SQLite3Inspector().get_tables()

    assert len(tables) == 2
    assert tables[0].name == "User"
    assert tables[0].columns[0].name == "user_id"
    assert tables[0].columns[0].data_type == "INTEGER"
    assert tables[0].columns[0].column_constraints == [
        c.PrimaryKeyConstraint("User", "user_id")
    ]
    assert tables[0].columns[1].name == "user_name"
    assert tables[0].columns[1].data_type == "TEXT"
    assert tables[0].columns[1].column_constraints == [
        c.NotNullConstraint("User", "user_name")
    ]

    assert tables[1].name == "Post"
    assert tables[1].columns[0].name == "post_id"
    assert tables[1].columns[0].data_type == "INTEGER"
    assert tables[1].columns[0].column_constraints == [
        c.PrimaryKeyConstraint("Post", "post_id")
    ]
    assert tables[1].columns[1].name == "post_name"
    assert tables[1].columns[1].data_type == "TEXT"
    assert tables[1].columns[1].column_constraints == [
        c.NotNullConstraint("Post", "post_name")
    ]
    assert tables[1].columns[2].name == "post_content"
    assert tables[1].columns[2].data_type == "TEXT"
    assert tables[1].columns[2].column_constraints == [
        c.NotNullConstraint("Post", "post_content")
    ]



import os

import py  # type: ignore
import pytest

from skibidi_orm.migration_engine.db_config.base_config import BaseDbConfig
from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
from skibidi_orm.migration_engine.studio.utils.db_config_dynamic_import import (
    db_config_dynamic_import,
)
import pathlib


def write_temp_schema_file(file_path: str, file_content: str):
    with open(file_path, "w") as f:
        f.write(file_content)


def test_db_config_dynamic_import_normal(tmp_path: pathlib.Path):
    # schema_file must be a different file in every test case.
    # Otherwise, import will cache the file and return the same instance.
    schema_file = tmp_path.joinpath("schema_test_db_config_dynamic_import_normal.py")
    schema_file.write_text(
        """from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
SQLite3Config(db_path="some/path/to/db.db")"""
    )
    str(schema_file)
    db_config = db_config_dynamic_import(schema_file_path=str(schema_file))

    assert isinstance(db_config, SQLite3Config)
    assert db_config.db_path == "some/path/to/db.db"
    assert BaseDbConfig.get_instance() == db_config


def test_db_config_dynamic_import_normal2(tmp_path: pathlib.Path):
    schema_file = tmp_path.joinpath("schema_test_db_config_dynamic_import_normal2.py")
    schema_file.write_text(
        """from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
SQLite3Config(db_path="some/path/to/db.db")"""
    )
    db_config = db_config_dynamic_import(schema_file_path=str(schema_file))

    assert isinstance(db_config, SQLite3Config)
    assert db_config.db_path == "some/path/to/db.db"
    assert BaseDbConfig.get_instance() == db_config


def test_db_config_dynamic_import_file_not_exists():
    schema_file = os.getcwd() + "/tmp/not_existant.py"
    with pytest.raises(FileNotFoundError):
        db_config_dynamic_import(schema_file_path=schema_file)


def test_db_config_dynamic_import_schema_path_is_not_file():
    schema_file = os.getcwd()
    with pytest.raises(TypeError):
        db_config_dynamic_import(schema_file_path=schema_file)



import sqlite3
from skibidi_orm.exceptions.db_mutator_exceptions import AmbigiousDeleteRowError
from skibidi_orm.migration_engine.data_mutator.base_data_mutator import (
    BaseDataMutator,
    DeleteRowPk,
    InsertRowColumn,
)
from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config


class SQLite3DataMutator(BaseDataMutator):

    def __init__(self):
        self.config = SQLite3Config.get_instance()

    def insert_row(self, table_name: str, row: list[InsertRowColumn]):
        """Insert a row in the table."""
        query = f"INSERT INTO {table_name} ({', '.join([col.name for col in row])}) VALUES ({', '.join(['?' for _ in row])})"
        values = [col.value for col in row]
        self._sqlite_execute(query, values)

    def delete_row(self, table_name: str, pks: list[DeleteRowPk]):
        """Delete a row in the table. Row identified by primary key subset."""
        deleted_rows = self._pre_select_to_delete_rows(table_name, pks)
        if len(deleted_rows) > 1:
            raise AmbigiousDeleteRowError(
                f"More than one row found with primary key subset: {pks}"
            )
        query = f"DELETE FROM {table_name} WHERE {' AND '.join([f'{pk.name} = ?' for pk in pks])}"
        values = [pk.value for pk in pks]
        self._sqlite_execute(query, values)

    def _pre_select_to_delete_rows(self, table_name: str, pks: list[DeleteRowPk]):
        """Select rows to be deleted. Used to check if the deletion is ambigious."""
        query = f"SELECT * FROM {table_name} WHERE {' AND '.join([f'{pk.name} = ?' for pk in pks])}"
        values = [pk.value for pk in pks]
        return self._sqlite_execute(query, values)

    def _sqlite_execute(self, query: str, values: list[str]):
        """
        Execute a query in the SQLite3 database, rutrns its result.
        """

        db_path = self.config.db_path
        with sqlite3.connect(db_path) as conn:
            cursor = conn.cursor()
            cursor.execute(query, values)
            conn.commit()
            data = cursor.fetchall()
            cursor.close()
        return data

    def raw_query(self, query: str):
        """Execute a raw sql query in the database."""
        return self._sqlite_execute(query, [])

    def get_rows(self, table_name: str, limit: int = 100, offset: int = 0):
        """Get paginated rows from the table."""
        query = f"SELECT * FROM {table_name} LIMIT {limit} OFFSET {offset}"
        return self._sqlite_execute(query, [])



from functools import wraps
import sqlite3
from typing import Any, Callable
from colorama import Style
import psycopg2
import pytest

from skibidi_orm.migration_engine.db_config.base_config import (
    BaseDbConfig,
)
import py  # type: ignore


@pytest.fixture(autouse=True)
def reset_config_singleton(monkeypatch: pytest.MonkeyPatch):
    # monkeypatch.setattr(ConfigSingleton, "_instances", {})
    # monkeypatch.setattr(BaseDbConfig, "_BaseDbConfig__instances_count", 0)
    monkeypatch.setattr(BaseDbConfig, "_BaseDbConfig__instance", None)


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item: Any) -> Any:
    """Print tests' docstring when running the test."""
    outcome: Any = yield
    report = outcome.get_result()
    test_fn = item.obj
    docstring = getattr(test_fn, "__doc__")
    if docstring:
        docstring = docstring.strip()
        report.nodeid = Style.DIM + docstring + Style.RESET_ALL


def execute_sqlite3_commands(db_path: str, commands: list[str]):
    """Executes the given commands in a SQLite DB living under db_path"""
    with sqlite3.connect(db_path) as conn:
        cursor = conn.cursor()
        for command in commands:
            cursor.execute(command)
        conn.commit()
        cursor.close()


@pytest.fixture
def make_database(request: pytest.FixtureRequest, tmpdir: py.path.local):
    sql_commands = (
        request.__getattribute__("param") if hasattr(request, "param") else None
    )
    p = tmpdir.join("temp_db.db")  # type: ignore
    p.write("")  # type: ignore
    if sql_commands:
        execute_sqlite3_commands(
            p.strpath,  # type: ignore
            sql_commands,
        )
    yield p
    tmpdir.remove()


def create_postgres_database(
    db_name: str,
    db_user: str,
    db_password: str,
    db_host: str,
    db_port: int,
    queries: list[str],
):
    connection = psycopg2.connect(
        user=db_user,
        password=db_password,
        host=db_host,
        port=db_port,
        database=db_name,
    )
    cursor = connection.cursor()
    for query in queries:
        cursor.execute(query)
    connection.commit()
    cursor.close()
    connection.close()


def clear_postgres_database(
    db_name: str,
    db_user: str,
    db_password: str,
    db_host: str,
    db_port: int,
):
    connection = psycopg2.connect(
        user=db_user,
        password=db_password,
        host=db_host,
        port=db_port,
        database=db_name,
    )
    cursor = connection.cursor()
    cursor.execute("DROP SCHEMA public CASCADE;")
    cursor.execute("CREATE SCHEMA public;")
    connection.commit()
    cursor.close()
    connection.close()


def postgres_db_fixture(
    db_name: str,
    db_user: str,
    db_password: str,
    db_host: str,
    db_port: int,
    queries: list[str],
):
    def decorator(func: Callable[..., Any]):
        @wraps(func)
        def wrapper():
            clear_postgres_database(db_name, db_user, db_password, db_host, db_port)
            create_postgres_database(
                db_name, db_user, db_password, db_host, db_port, queries
            )
            func()
            # clear_postgres_database(db_name, db_user, db_password, db_host, db_port)

        return wrapper

    return decorator



import sqlite3
import pytest

from skibidi_orm.exceptions.db_mutator_exceptions import AmbigiousDeleteRowError
from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    NotNullConstraint,
    PrimaryKeyConstraint,
)
from skibidi_orm.migration_engine.adapters.database_objects.sqlite3_typing import (
    SQLite3Typing,
)
from skibidi_orm.migration_engine.data_mutator.base_data_mutator import (
    DeleteRowPk,
    InsertRowColumn,
)
from skibidi_orm.migration_engine.data_mutator.sqlite3_data_mutatorr import (
    SQLite3DataMutator,
)
from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
from skibidi_orm.migration_engine.db_inspectors.sqlite3_inspector import (
    SQLite3Inspector,
)
from ..sql_data import SQLite3InsertData, SQLite3TablesData


@pytest.mark.parametrize(
    "make_database", [[*SQLite3TablesData.sql_simple_db]], indirect=True
)
def test_insert_row(make_database: str):
    SQLite3Config(db_path=make_database)
    db_seeder = SQLite3DataMutator()
    db_seeder.insert_row(
        table_name="users",
        row=[
            InsertRowColumn(name="user_id", value=str(1)),
            InsertRowColumn(name="username", value="test"),
        ],
    )
    db_inspector = SQLite3Inspector()
    tables = db_inspector.get_tables()

    assert tables[0] == SQLite3Typing.Table(
        name="users",
        columns=[
            SQLite3Typing.Column(
                name="user_id",
                data_type="INTEGER",
                column_constraints=[
                    PrimaryKeyConstraint(table_name="users", column_name="user_id"),
                ],
            ),
            SQLite3Typing.Column(
                name="username",
                data_type="TEXT",
                column_constraints=[
                    NotNullConstraint(table_name="users", column_name="username"),
                ],
            ),
        ],
    )


@pytest.mark.parametrize(
    "make_database", [[*SQLite3TablesData.sql_simple_db]], indirect=True
)
def test_delete_row_one_pk(make_database: str):
    config = SQLite3Config(db_path=make_database)
    db_seeder = SQLite3DataMutator()

    db_seeder.insert_row(
        table_name="users",
        row=[
            InsertRowColumn(name="user_id", value=str(1)),
            InsertRowColumn(name="username", value="test"),
        ],
    )

    with sqlite3.connect(config.db_path) as conn:
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM users")
        data = cursor.fetchall()
        cursor.close()

    assert len(data) == 1

    db_seeder.delete_row(
        table_name="users",
        pks=[
            DeleteRowPk(name="user_id", value=str(1)),
        ],
    )

    with sqlite3.connect(config.db_path) as conn:
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM users")
        data = cursor.fetchall()
        cursor.close()

    assert len(data) == 0


@pytest.mark.parametrize(
    "make_database", [[*SQLite3TablesData.sql_double_pk_db]], indirect=True
)
def test_delete_row_two_pk(make_database: str):
    config = SQLite3Config(db_path=make_database)
    db_seeder = SQLite3DataMutator()
    db_seeder.insert_row(
        table_name="Orders",
        row=[
            InsertRowColumn(name="order_id", value=str(1)),
            InsertRowColumn(name="product_id", value=str(2)),
        ],
    )

    with sqlite3.connect(config.db_path) as conn:
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM Orders")
        data = cursor.fetchall()
        cursor.close()

    assert len(data) == 1

    db_seeder.delete_row(
        table_name="Orders",
        pks=[
            DeleteRowPk(name="order_id", value=str(1)),
            DeleteRowPk(name="product_id", value=str(2)),
        ],
    )

    with sqlite3.connect(config.db_path) as conn:
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM Orders")
        data = cursor.fetchall()
        cursor.close()

    assert len(data) == 0


@pytest.mark.parametrize(
    "make_database", [[*SQLite3TablesData.sql_double_pk_db]], indirect=True
)
def test_delete_row_table_two_pk_one_pk_given_ok(make_database: str):
    config = SQLite3Config(db_path=make_database)
    db_seeder = SQLite3DataMutator()
    db_seeder.insert_row(
        table_name="Orders",
        row=[
            InsertRowColumn(name="order_id", value=str(1)),
            InsertRowColumn(name="product_id", value=str(1)),
        ],
    )
    db_seeder.insert_row(
        table_name="Orders",
        row=[
            InsertRowColumn(name="order_id", value=str(1)),
            InsertRowColumn(name="product_id", value=str(2)),
        ],
    )

    with sqlite3.connect(config.db_path) as conn:
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM Orders")
        data = cursor.fetchall()
        cursor.close()

    assert len(data) == 2

    db_seeder.delete_row(
        table_name="Orders",
        pks=[
            DeleteRowPk(name="product_id", value=str(2)),
        ],
    )

    with sqlite3.connect(config.db_path) as conn:
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM Orders")
        data = cursor.fetchall()
        cursor.close()

    assert len(data) == 1
    assert data[0] == (1, 1)


@pytest.mark.parametrize(
    "make_database", [[*SQLite3TablesData.sql_double_pk_db]], indirect=True
)
def test_delete_row_table_two_pk_two_pk_given_ok(make_database: str):
    config = SQLite3Config(db_path=make_database)
    db_seeder = SQLite3DataMutator()
    db_seeder.insert_row(
        table_name="Orders",
        row=[
            InsertRowColumn(name="order_id", value=str(1)),
            InsertRowColumn(name="product_id", value=str(1)),
        ],
    )
    db_seeder.insert_row(
        table_name="Orders",
        row=[
            InsertRowColumn(name="order_id", value=str(1)),
            InsertRowColumn(name="product_id", value=str(2)),
        ],
    )

    with sqlite3.connect(config.db_path) as conn:
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM Orders")
        data = cursor.fetchall()
        cursor.close()

    assert len(data) == 2

    db_seeder.delete_row(
        table_name="Orders",
        pks=[
            DeleteRowPk(name="order_id", value=str(1)),
            DeleteRowPk(name="product_id", value=str(2)),
        ],
    )

    with sqlite3.connect(config.db_path) as conn:
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM Orders")
        data = cursor.fetchall()
        cursor.close()

    assert len(data) == 1
    assert data[0] == (1, 1)


@pytest.mark.parametrize(
    "make_database", [[*SQLite3TablesData.sql_double_pk_db]], indirect=True
)
def test_delete_row_table_two_pk_one_pk_given_ambigious(make_database: str):
    config = SQLite3Config(db_path=make_database)
    db_seeder = SQLite3DataMutator()
    db_seeder.insert_row(
        table_name="Orders",
        row=[
            InsertRowColumn(name="order_id", value=str(1)),
            InsertRowColumn(name="product_id", value=str(1)),
        ],
    )
    db_seeder.insert_row(
        table_name="Orders",
        row=[
            InsertRowColumn(name="order_id", value=str(1)),
            InsertRowColumn(name="product_id", value=str(2)),
        ],
    )

    with sqlite3.connect(config.db_path) as conn:
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM Orders")
        data = cursor.fetchall()
        cursor.close()

    assert len(data) == 2

    with pytest.raises(AmbigiousDeleteRowError):
        db_seeder.delete_row(
            table_name="Orders",
            pks=[
                DeleteRowPk(name="order_id", value=str(1)),
            ],
        )


@pytest.mark.parametrize(
    "make_database",
    [[*SQLite3TablesData.sql_simple_db, *SQLite3InsertData.sql_simple_insert]],
    indirect=True,
)
def test_raw_query(make_database: str):
    SQLite3Config(db_path=make_database)
    db_mutator = SQLite3DataMutator()
    data = db_mutator.raw_query("SELECT * FROM users;")
    assert data == [(1, "test1"), (2, "test2"), (3, "test3")]


@pytest.mark.parametrize(
    "make_database",
    [[*SQLite3TablesData.sql_simple_db, *SQLite3InsertData.sql_simple_insert]],
    indirect=True,
)
def test_get_rows_normal(make_database: str):
    SQLite3Config(db_path=make_database)
    db_mutator = SQLite3DataMutator()
    data = db_mutator.get_rows("users")
    assert data == [(1, "test1"), (2, "test2"), (3, "test3")]


@pytest.mark.parametrize(
    "make_database",
    [[*SQLite3TablesData.sql_simple_db, *SQLite3InsertData.sql_simple_insert]],
    indirect=True,
)
def test_get_rows_offset(make_database: str):
    SQLite3Config(db_path=make_database)
    db_mutator = SQLite3DataMutator()
    data = db_mutator.get_rows("users", offset=1)
    assert data == [(2, "test2"), (3, "test3")]
    data = db_mutator.get_rows("users", offset=2)
    assert data == [(3, "test3")]
    data = db_mutator.get_rows("users", offset=3)
    assert data == []


@pytest.mark.parametrize(
    "make_database",
    [[*SQLite3TablesData.sql_simple_db, *SQLite3InsertData.sql_simple_insert]],
    indirect=True,
)
def test_get_rows_limit(make_database: str):
    SQLite3Config(db_path=make_database)
    db_mutator = SQLite3DataMutator()
    data = db_mutator.get_rows("users", limit=1)
    assert data == [(1, "test1")]
    data = db_mutator.get_rows("users", limit=2)
    assert data == [(1, "test1"), (2, "test2")]
    data = db_mutator.get_rows("users", limit=3)
    assert data == [(1, "test1"), (2, "test2"), (3, "test3")]
    data = db_mutator.get_rows("users", limit=4)
    assert data == [(1, "test1"), (2, "test2"), (3, "test3")]



class SQLite3TablesData:
    sql_table1 = """
        CREATE TABLE table1 (
            id INTEGER PRIMARY KEY,
            name TEXT NOT NULL
        );
    """
    sql_table2 = """
        CREATE TABLE table2 (
            id INTEGER PRIMARY KEY,
            name TEXT NOT NULL
        );
    """

    sql_table_primary_key_not_null = """
        CREATE TABLE table_primary_key_not_null (
            id INTEGER PRIMARY KEY NOT NULL
        );
    """

    sql_schema_with_fks = [
        """
        CREATE TABLE users (
            user_id INTEGER PRIMARY KEY,
            username TEXT NOT NULL UNIQUE,
            email TEXT NOT NULL UNIQUE,
            password_hash TEXT NOT NULL,
            registration_date TIMESTAMP DEFAULT CURRENT_TIMESTAMP
        );
    """,
        """
        CREATE TABLE posts (
            post_id INTEGER PRIMARY KEY,
            user_id INTEGER NOT NULL,
            title TEXT NOT NULL,
            content TEXT NOT NULL,
            post_date TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
            FOREIGN KEY (user_id) REFERENCES users(user_id)
        );
    """,
        """
        CREATE TABLE comments (
            comment_id INTEGER PRIMARY KEY,
            username TEXT NOT NULL,
            user_idd INTEGER NOT NULL,
            post_id INTEGER NOT NULL,
            comment_text TEXT NOT NULL,
            comment_date TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
            FOREIGN KEY (user_idd, username) REFERENCES users(user_id, username),
            FOREIGN KEY (post_id) REFERENCES posts(post_id)
        );
    """,
    ]
    sql_table_user = """
        CREATE TABLE User (
            user_id INTEGER PRIMARY KEY,
            user_name TEXT NOT NULL
        );
    """
    sql_table_post = """
        CREATE TABLE Post (
            post_id INTEGER PRIMARY KEY,
            post_name TEXT NOT NULL
        );
    """

    sql_table_comment = """
        CREATE TABLE Comment (
            comment_id INTEGER PRIMARY KEY,
            comment_name TEXT NOT NULL
        );
    """
    sql_table_primary_key_not_null = """
        CREATE TABLE table_primary_key_not_null (
            id INTEGER PRIMARY KEY NOT NULL
        );
    """

    sql_simple_db = [
        """
        CREATE TABLE users (
            user_id INTEGER PRIMARY KEY,
            username TEXT NOT NULL
        );
    """
    ]
    sql_double_pk_db = [
        """
        CREATE TABLE Orders (
            order_id INTEGER,
            product_id INTEGER,
            PRIMARY KEY (order_id, product_id)
        );
    """
    ]


class PostgresTablesData:
    sql_table1 = SQLite3TablesData.sql_table1
    sql_table2 = SQLite3TablesData.sql_table2
    sql_table_primary_key_not_null = SQLite3TablesData.sql_table_primary_key_not_null
    sql_schema_with_fks = SQLite3TablesData.sql_schema_with_fks
    sql_table_user = SQLite3TablesData.sql_table_user
    sql_table_post = SQLite3TablesData.sql_table_post
    sql_table_comment = SQLite3TablesData.sql_table_comment
    sql_table_primary_key_not_null = SQLite3TablesData.sql_table_primary_key_not_null
    sql_simple_db = SQLite3TablesData.sql_simple_db
    sql_double_pk_db = SQLite3TablesData.sql_double_pk_db
    SQL_TABLE_MANY_COLUMNS1 = """
        CREATE TABLE table_many_columns1 (
            primary_key INTEGER PRIMARY KEY,
            text_not_null TEXT NOT NULL,
            text_nullabe TEXT,
            integer_not_null INTEGER NOT NULL,
            integer_nullable INTEGER,
            date_not_null DATE NOT NULL,
            date_nullable DATE
        );
"""
    SQL_TABLE_DIFFICULT_TYPES = """
        CREATE TABLE table_difficult_types (
            big_primary_key_alias serial8 PRIMARY KEY,
            bit_varying_alias VARBIT,
            bit_varying_alias_arg VARBIT(20),
            double_precision DOUBLE PRECISION,
            timestamp TIMESTAMP,
            timestamp_arg TIMESTAMP(2),
            timestamp_without_time_zone TIMESTAMP WITHOUT TIME ZONE,
            timestamp_without_time_zone_arg TIMESTAMP(3) WITHOUT TIME ZONE,
            timestamp_with_time_zone_alias TIMESTAMPTZ,
            timestamp_with_time_zone_arg_alias TIMESTAMPTZ(3),
            charachter_varying CHARACTER VARYING,
            charachter_varying_arg CHARACTER VARYING(40),
            charachter_varying_alias VARCHAR,
            charachter_varying_alias_args VARCHAR(40)
        );
"""
    SQL_TABLE_DIFFERECT_CONSTRAINTS = """
        CREATE TABLE table_different_constraints (
            primary_key INTEGER PRIMARY KEY,
            integer_not_nullable INTEGER NOT NULL,
            text_nullable TEXT,
            unique_column INTEGER UNIQUE,
            unique_not_nullable INTEGER NOT NULL UNIQUE,
            check_other_column INTEGER CHECK (integer_not_nullable > 100),
            default_column INTEGER DEFAULT 1,
            not_null_unique_check_column INTEGER NOT NULL UNIQUE CHECK (not_null_unique_check_column > 100),
            not_null_unique INTEGER NOT NULL UNIQUE,
            not_null_default INTEGER NOT NULL DEFAULT 1,
            not_null_unique_check_default INTEGER NOT NULL UNIQUE DEFAULT 1 CHECK (not_null_unique_check_default > 100)
        );
"""
    SQL_TABLE_SERIALS = """
        CREATE TABLE table_serials (
            serial_pk SERIAL PRIMARY KEY,
            bigserial BIGSERIAL
        );
"""
    SQL_TABLE_DEFAULTS = """
        CREATE TABLE table_defaults (
            primary_key INTEGER PRIMARY KEY,
            integer_default INTEGER DEFAULT 1,
            text_default TEXT DEFAULT 'default',
            date_default DATE DEFAULT CURRENT_DATE,
            timestamp_default TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
            timestamp_default_lowercase TIMESTAMP DEFAULT current_timestamp
        );
    """
    SQL_TABLE_SIMPLE_FOREIGN_KEYS = [
        """
        CREATE TABLE authors (
            id SERIAL PRIMARY KEY,
            name VARCHAR(100) NOT NULL
        );
        """,
        """
        CREATE TABLE books (
            id SERIAL PRIMARY KEY,
            title VARCHAR(255) NOT NULL,
            year INT,
            author_id_foreign_key INT,
            FOREIGN KEY (author_id_foreign_key) REFERENCES authors (id)
        );
        """,
    ]
    SQL_TABLE_FKS_MORE_COMPLEX = [
        """
        CREATE TABLE departments (
            department_id SERIAL PRIMARY KEY
        );
        """,
        """
        CREATE TABLE employees (
            employee_id SERIAL PRIMARY KEY,
            department_id INTEGER NOT NULL,
            CONSTRAINT fk_department
                FOREIGN KEY (department_id)
                REFERENCES Departments(department_id)
        );
        """,
        """
        CREATE TABLE projects (
            project_id SERIAL PRIMARY KEY,
            project_name VARCHAR(255) NOT NULL,
            lead_employee_id INTEGER NOT NULL,
            support_department_id INTEGER NOT NULL,
            FOREIGN KEY (lead_employee_id) REFERENCES Employees(employee_id),
            FOREIGN KEY (support_department_id) REFERENCES Departments(department_id)
        );
        """,
    ]


class SQLite3InsertData:
    sql_simple_insert = [
        """
        INSERT INTO users (user_id, username) VALUES
        (1, 'test1'),
        (2, 'test2'),
        (3, 'test3')
    ;
    """
    ]


class PostgresInsertData:
    sql_simple_insert = SQLite3InsertData.sql_simple_insert



from skibidi_orm.migration_engine.db_config.base_config import (
    BaseDbConfig,
)


class SQLite3Config(BaseDbConfig):
    """
    Configuration class for SQLite3 database.
    Instantiating it means choosing SQLite3 as the database.
    """

    def __init__(self, db_path: str):
        self.__db_path = db_path

    @property
    def db_path(self) -> str:
        return self.__db_path



import importlib

from fastapi.testclient import TestClient
import pytest
from skibidi_orm.migration_engine.data_mutator.sqlite3_data_mutatorr import (
    SQLite3DataMutator,
)
from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
from skibidi_orm.migration_engine.db_inspectors.sqlite3_inspector import (
    SQLite3Inspector,
)
from skibidi_orm.migration_engine.studio.server import app
import pathlib
from ..sql_data import SQLite3TablesData

client = TestClient(app)


@pytest.mark.parametrize(
    "make_database", [[*SQLite3TablesData.sql_simple_db]], indirect=True
)
def test_GET_db(
    monkeypatch: pytest.MonkeyPatch, tmp_path: pathlib.Path, make_database: str
):
    import skibidi_orm.migration_engine.studio.server  # type: ignore

    importlib.import_module("skibidi_orm.migration_engine.studio.server")

    SQLite3Config(db_path=make_database)
    monkeypatch.setattr(
        "skibidi_orm.migration_engine.studio.server.db_inspector",
        SQLite3Inspector(),
    )
    response = client.get("/db")
    assert response.status_code == 200
    assert sorted(response.json()) == sorted(
        {
            "tables": [
                {
                    "name": "users",
                    "columns": [
                        {
                            "constraints": [
                                {
                                    "constraint_type": "PRIMARY KEY",
                                    "column_name": "username",
                                    "table_name": "users",
                                }
                            ],
                            "data_type": "INTEGER",
                            "name": "user_id",
                        },
                        {
                            "data_type": "TEXT",
                            "name": "username",
                            "constraints": [
                                {
                                    "constraint_type": "NOT NULL",
                                    "column_name": "username",
                                    "table_name": "users",
                                }
                            ],
                        },
                    ],
                }
            ]
        }
    )


@pytest.mark.parametrize(
    "make_database", [[*SQLite3TablesData.sql_simple_db]], indirect=True
)
def test_POST_route_db_table_name_row_correct(
    monkeypatch: pytest.MonkeyPatch, make_database: str
):
    import skibidi_orm.migration_engine.studio.server  # type: ignore

    importlib.import_module("skibidi_orm.migration_engine.studio.server")

    SQLite3Config(db_path=make_database)
    monkeypatch.setattr(
        "skibidi_orm.migration_engine.studio.server.db_inspector",
        SQLite3Inspector(),
    )
    monkeypatch.setattr(
        "skibidi_orm.migration_engine.studio.server.db_mutator",
        SQLite3DataMutator(),
    )
    response = client.post(
        "/db/users/row",
        json={
            "row": [
                {"name": "user_id", "value": "1"},
                {"name": "username", "value": "test"},
            ]
        },
    )
    assert response.status_code == 200
    assert response.json() == {"message": "Row inserted successfully."}



from typing import Callable
from typer.testing import CliRunner
import typer
import pytest
from skibidi_orm.cli.run import app
import importlib
import py  # type: ignore

runner = CliRunner()


@pytest.fixture
def mock_input_id(monkeypatch: pytest.MonkeyPatch):
    def mock_input(prompt: pytest.MonkeyPatch):
        return "1"

    monkeypatch.setattr(typer, "prompt", mock_input)


def test_list_migrations():
    result = runner.invoke(app, ["list-migrations"])
    assert result.exit_code == 0
    assert "Listing all migrations" in result.stdout


def test_go():
    result = runner.invoke(app, ["go", "1"])
    assert result.exit_code == 0
    assert "Going to migration with ID: 1" in result.stdout


def test_go_no_migration_id(mock_input_id: pytest.MonkeyPatch):
    result = runner.invoke(app, ["go"])
    assert result.exit_code == 0
    assert "Going to migration with ID: 1" in result.stdout


def test_studio_no_options_too_many_schemas(tmpdir: py.path.local):  # type: ignore
    file_path = tmpdir.join("schema.py")  # type: ignore
    file_path2 = tmpdir.mkdir("other").join("schema.py")  # type: ignore
    file_path.write("")  # type: ignore
    file_path2.write("")  # type: ignore
    result = runner.invoke(app, ["studio"])
    assert result.exit_code == 1
    assert (
        "Multiple schema files found. Please specify the schema file to use: --schema-file <PATH>"
        in result.stdout
    )
    tmpdir.remove()


def test_studio_one_schema_option(monkeypatch: pytest.MonkeyPatch, tmpdir: py.path.local):  # type: ignore
    import skibidi_orm.migration_engine.studio.server  # type: ignore

    def mock_func(schema_file: str):
        print("Success test", end="")

    monkeypatch.setattr(
        "skibidi_orm.migration_engine.studio.server.run_server", mock_func
    )
    importlib.reload(skibidi_orm.cli.run)  # type: ignore

    file_path = tmpdir.join("schema.py")  # type: ignore
    file_path.write("")  # type: ignore

    result = runner.invoke(
        app, ["studio", "--schema-file", file_path.strpath]  # type: ignore
    )

    assert "Success test" == result.stdout
    tmpdir.remove()



import pytest

from skibidi_orm.migration_engine.adapters.database_objects.sqlite3_typing import SQLite3Typing
from skibidi_orm.migration_engine.converters.sqlite3.tables import SQLite3TableOperationConverter
from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    CheckConstraint,
    Constraint,
    ConstraintType,
    ForeignKeyConstraint,
    PrimaryKeyConstraint,
    UniqueConstraint,
    NotNullConstraint,
)
from skibidi_orm.migration_engine.operations.table_operations import (
    CreateTableOperation,
    DeleteTableOperation,
    RenameTableOperation,
)
from itertools import chain


simple_table_no_constraints = SQLite3Typing.Table(
    "users", columns=[SQLite3Typing.Column("name", "TEXT")]
)


complex_table_no_constraints = SQLite3Typing.Table(
    "users",
    columns=[
        SQLite3Typing.Column("name", "TEXT"),
        SQLite3Typing.Column("age", "INTEGER"),
        SQLite3Typing.Column("email", "TEXT"),
        SQLite3Typing.Column("active", "BLOB"),
    ],
)

complex_table_with_constraints = SQLite3Typing.Table(
    "admin_users",
    columns=[
        SQLite3Typing.Column(
            "user_id",
            "INTEGER",
            column_constraints=[
                PrimaryKeyConstraint("users", "user_id"),
                NotNullConstraint("users", "user_id"),
            ],
        ),
        SQLite3Typing.Column("name", "TEXT"),
        SQLite3Typing.Column(
            "email", "TEXT", column_constraints=[UniqueConstraint("users", "email")]
        ),
        SQLite3Typing.Column("active", "BLOB", column_constraints=list()),
        SQLite3Typing.Column(
            "age",
            "INTEGER",
            column_constraints=[CheckConstraint("users", "age", "> 18")],
        ),
    ],
    foreign_keys={
        ForeignKeyConstraint(
            "admin_users", "users", {"active": "active", "name": "name"}
        )
    },
)


@pytest.fixture
def non_random_constraint_order(monkeypatch: pytest.MonkeyPatch):
    """Since the actual split_constraints method works on sets, so the
    order of constraints is not guaranteed, we have to mock it using lists"""

    def split_using_lists(
        table: SQLite3Typing.Table,
    ) -> tuple[list[Constraint], list[Constraint]]:
        all_constraints = list(
            chain.from_iterable(column.column_constraints for column in table.columns)
        )

        # TODO: make this dependent on the implementation of the TableOperationConverter instead of
        #  doing it here
        constraints_at_end = list(
            filter(
                lambda c: c.constraint_type
                not in [
                    ConstraintType.PRIMARY_KEY,
                    ConstraintType.UNIQUE,
                    ConstraintType.NOT_NULL,
                ],
                all_constraints,
            )
        )

        constraints_at_columns = [
            constraint
            for constraint in all_constraints
            if constraint not in constraints_at_end
        ]

        return list(constraints_at_end), list(constraints_at_columns)

    monkeypatch.setattr(
        SQLite3TableOperationConverter, "split_constraints", split_using_lists
    )


def test_create_table_conversion_simple():
    """Create a simple table"""
    operation = CreateTableOperation(simple_table_no_constraints)
    assert (
        SQLite3TableOperationConverter.convert_table_operation_to_SQL(operation)
        == "CREATE TABLE users (name TEXT);"
    )


def test_create_table_conversion_complex():
    """Create a complex table"""
    operation = CreateTableOperation(complex_table_no_constraints)
    assert (
        SQLite3TableOperationConverter.convert_table_operation_to_SQL(operation)
        == "CREATE TABLE users (name TEXT, age INTEGER, email TEXT, active BLOB);"
    )


@pytest.mark.usefixtures("non_random_constraint_order")
def test_create_table_conversion_complex_with_constraints():
    """Create a complex table with multiple constraints"""
    operation = CreateTableOperation(complex_table_with_constraints)
    assert (
        SQLite3TableOperationConverter.convert_table_operation_to_SQL(operation)
        == "CREATE TABLE admin_users (user_id INTEGER PRIMARY KEY NOT NULL, name TEXT, "
        "email TEXT UNIQUE, active BLOB, age INTEGER, CHECK (age > 18), FOREIGN KEY (active, name) REFERENCES users ("
           "active, name));"
    )


def test_delete_table_conversion():
    """Delete a simple table"""
    operation = DeleteTableOperation(simple_table_no_constraints)
    assert (
        SQLite3TableOperationConverter.convert_table_operation_to_SQL(operation)
        == "DROP TABLE users;"
    )


def test_rename_table_conversion_simple():
    """Rename a table"""
    operation = RenameTableOperation(simple_table_no_constraints, "users2")
    assert (
        SQLite3TableOperationConverter.convert_table_operation_to_SQL(operation)
        == "DROP TABLE users; CREATE TABLE users2 (name TEXT);"
    )



from __future__ import annotations
from abc import ABC, abstractmethod

from skibidi_orm.exceptions.operations import IrreversibleOperationError
from skibidi_orm.migration_engine.operations.operation_type import OperationType
from skibidi_orm.migration_engine.adapters.base_adapter import BaseTable, BaseColumn
from typing import Any
from dataclasses import dataclass, field


@dataclass(frozen=True)
class TableOperation(ABC):
    """Base class for table operations"""

    operation_type: OperationType
    table: BaseTable[BaseColumn[Any]]
    is_reversible: bool

    @abstractmethod
    def reverse(self) -> TableOperation:
        """Generates a table operation that reverses
        the calling instance if possible. Else, throws an exception"""


@dataclass(frozen=True)
class CreateTableOperation(TableOperation):
    """Class for creating a table"""

    operation_type: OperationType = field(init=False, default=OperationType.CREATE)
    is_reversible: bool = field(init=False, default=True)

    def reverse(self) -> TableOperation:
        return DeleteTableOperation(table=self.table)

    def __str__(self) -> str:
        return f"Create Table {self.table.name} with columns: {', '.join([col.name for col in self.table.columns])}"


@dataclass(frozen=True)
class DeleteTableOperation(TableOperation):
    """Class for deleting a table"""

    operation_type: OperationType = field(init=False, default=OperationType.DELETE)
    is_reversible: bool = field(init=False, default=False)

    def reverse(self) -> TableOperation:
        raise IrreversibleOperationError(
            f"Reversing a {self.__class__.__name__} is currently not supported."
        )

    def __str__(self) -> str:
        return f"Delete Table {self.table.name}"


@dataclass(frozen=True)
class RenameTableOperation(TableOperation):
    """Class for renaming a table"""

    operation_type: OperationType = field(init=False, default=OperationType.RENAME)
    is_reversible: bool = field(init=False, default=True)
    new_name: str

    def reverse(self) -> TableOperation:
        # todo: make sure it works properly
        return RenameTableOperation(table=self.table, new_name=self.table.name)

    def __str__(self) -> str:
        return f"Rename Table {self.table.name} to {self.new_name}"



from skibidi_orm.migration_engine.converters.base.interfaces import (
    ColumnOperationSQLConverter,
    ConstraintSQLConverter,
    SQLConverter,
    TableOperationSQLConverter,
)
from skibidi_orm.migration_engine.converters.sqlite3.columns import (
    SQLite3ColumnOperationConverter,
)
from skibidi_orm.migration_engine.converters.sqlite3.constraints import (
    SQLite3ConstraintConverter,
)
from skibidi_orm.migration_engine.converters.sqlite3.tables import (
    SQLite3TableOperationConverter,
)


class SQLite3Converter(SQLConverter):
    """Class reponsible for converting operation and constraint
    objects to SQLite3 SQL strings"""

    @staticmethod
    def get_constraint_converter() -> type[ConstraintSQLConverter]:
        return SQLite3ConstraintConverter

    @staticmethod
    def get_table_operation_converter() -> type[TableOperationSQLConverter]:
        return SQLite3TableOperationConverter

    @staticmethod
    def get_column_operation_converter() -> type[ColumnOperationSQLConverter]:
        return SQLite3ColumnOperationConverter



from skibidi_orm.migration_engine.adapters.database_objects.sqlite3_typing import SQLite3Typing
from skibidi_orm.migration_engine.converters.base.interfaces import (
    TableOperationSQLConverter,
)
from skibidi_orm.migration_engine.converters.sqlite3.columns import (
    SQLite3ColumnOperationConverter,
)
from skibidi_orm.migration_engine.converters.sqlite3.constraints import (
    SQLite3ConstraintConverter,
)
from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    ColumnSpecificConstraint,
    ConstraintType,
)
from skibidi_orm.migration_engine.operations.operation_type import OperationType
from skibidi_orm.migration_engine.operations.table_operations import (
    DeleteTableOperation,
    RenameTableOperation,
    TableOperation,
    CreateTableOperation,
)
from skibidi_orm.exceptions.operations import UnsupportedOperationError
from typing import cast
from itertools import chain


class SQLite3TableOperationConverter(TableOperationSQLConverter):
    """Class responsible for converting table operation objects to raw SQLite3 SQL strings"""

    @staticmethod
    def convert_table_operation_to_SQL(operation: TableOperation) -> str:
        """Convert a given table operation to a SQLite3 SQL string"""
        if operation.operation_type == OperationType.CREATE:
            return (
                SQLite3TableOperationConverter._convert_create_table_operation_to_SQL(
                    cast(CreateTableOperation, operation)
                )
            )
        elif operation.operation_type == OperationType.DELETE:
            return SQLite3TableOperationConverter._convert_drop_table_operation_to_SQL(
                cast(DeleteTableOperation, operation)
            )
        elif operation.operation_type == OperationType.RENAME:
            return (
                SQLite3TableOperationConverter._convert_rename_table_operation_to_SQL(
                    cast(RenameTableOperation, operation)
                )
            )
        else:
            raise UnsupportedOperationError(
                "The given operation is not supported in SQLite3."
            )

    @staticmethod
    def _convert_create_table_operation_to_SQL(operation: CreateTableOperation) -> str:
        """Convert a given create table operation to a SQLite3 SQL string"""
        definition_string = f"CREATE TABLE {operation.table.name} "
        constraints_at_end, constraints_at_columns = (
            SQLite3TableOperationConverter.split_constraints(operation.table)
        )

        definition_string += "("
        for column in operation.table.columns:
            constraints_at_column = [
                constraint
                for constraint in column.column_constraints
                if constraint in constraints_at_columns
            ]
            definition_string += f"{SQLite3ColumnOperationConverter.convert_column_definition_to_SQL(
                    column, constraints_at_column)}, "

        for constraint in constraints_at_end:
            definition_string += (
                f"{SQLite3ConstraintConverter.convert_constraint_to_SQL(constraint)}, "
            )

        for key in operation.table.foreign_keys:
            definition_string += (
                f"{SQLite3ConstraintConverter.convert_constraint_to_SQL(key)}, "
            )

        return definition_string.strip(", ") + ");"

    @staticmethod
    def _convert_drop_table_operation_to_SQL(operation: DeleteTableOperation) -> str:
        """Convert a given drop table operation to a SQLite3 SQL string"""
        return f"DROP TABLE {operation.table.name};"

    @staticmethod
    def _convert_rename_table_operation_to_SQL(operation: RenameTableOperation) -> str:
        """Convert a given rename table operation to a SQLite3 SQL string. Since SQLite3 does not
        support renaming a table directly, a new table with the same columns has to be instantiated after
        deleting the original one."""
        delete_op = DeleteTableOperation(operation.table)
        create_op = CreateTableOperation(
            SQLite3Typing.Table(operation.new_name, operation.table.columns)
        )
        return (
            SQLite3TableOperationConverter.convert_table_operation_to_SQL(delete_op)
            + " "
            + SQLite3TableOperationConverter.convert_table_operation_to_SQL(create_op)
        )

    @staticmethod
    def split_constraints(
        table: SQLite3Typing.Table,
    ) -> tuple[set[ColumnSpecificConstraint], set[ColumnSpecificConstraint]]:
        """Split the constraints of a table into those that have to be added at the end of the
        table definition and those that have to be added in the definitions of their resective columns
        """
        all_constraints = set(
            chain.from_iterable(column.column_constraints for column in table.columns)
        )

        """These constraints have to be added at the end of the table definition, not by the
        columns they correspond to"""
        constraints_at_end = set(
            filter(
                lambda c: c.constraint_type
                not in [ConstraintType.PRIMARY_KEY, ConstraintType.UNIQUE, ConstraintType.NOT_NULL],
                all_constraints,
            )
        )

        # These constraints have to be added in the column definition
        constraints_at_columns = all_constraints - constraints_at_end

        return set(constraints_at_end), set(constraints_at_columns)



from abc import ABC, abstractmethod

from skibidi_orm.migration_engine.operations.table_operations import TableOperation
from skibidi_orm.migration_engine.operations.column_operations import ColumnOperation


class BaseSQLExecutor(ABC):

    @staticmethod
    @abstractmethod
    def execute_sql(sql: str) -> None:
        """Execute the given SQL query in the database"""
        pass

    @staticmethod
    @abstractmethod
    def execute_operations(operations: list[TableOperation | ColumnOperation]) -> None:
        """Execute the given operations in the database"""
        pass



from __future__ import annotations
from abc import abstractmethod, ABC
from skibidi_orm.migration_engine.operations.column_operations import ColumnOperation
from skibidi_orm.migration_engine.operations.table_operations import TableOperation
from skibidi_orm.migration_engine.adapters.database_objects.constraints import (
    Constraint,
)


class SQLConverter(ABC):
    """Class responsible for converting operation and constraint objects
    to raw SQL strings in a specific dialect. Each of the supported database
    providers should inherit a converter from this base class and implement its
    methods"""

    @staticmethod
    @abstractmethod
    def get_column_operation_converter() -> type[ColumnOperationSQLConverter]:
        """Return the corresponding column operation converter class"""

    @staticmethod
    @abstractmethod
    def get_table_operation_converter() -> type[TableOperationSQLConverter]:
        """Return the corresponsing table operation converter class"""

    @staticmethod
    @abstractmethod
    def get_constraint_converter() -> type[ConstraintSQLConverter]:
        """Return the corresponding constraint converter class"""

    @classmethod
    def convert_table_operation_to_SQL(cls, operation: TableOperation) -> str:
        """Convert a given table operation object to raw SQL in a specific dialect"""
        return cls.get_table_operation_converter().convert_table_operation_to_SQL(
            operation
        )

    @classmethod
    def convert_column_operation_to_SQL(cls, operation: ColumnOperation) -> str:
        """Convert a given column operation object to raw SQL in a specific dialect"""
        return cls.get_column_operation_converter().convert_column_operation_to_SQL(
            operation
        )

    @classmethod
    def _convert_constraint_to_SQL(cls, constraint: Constraint) -> str:
        """Convert a given constraint to raw SQL in a specific dialect"""
        return cls.get_constraint_converter().convert_constraint_to_SQL(constraint)


class ColumnOperationSQLConverter(ABC):
    """Class responsible for converting column operation objects to raw SQL strings
    in a specific dialect. Each of the supported database providers should inherit
    a converter from this base class and implement its methods"""

    @staticmethod
    @abstractmethod
    def convert_column_operation_to_SQL(operation: ColumnOperation) -> str:
        """Convert a given column operation object to raw SQL in a specific dialect"""


class TableOperationSQLConverter(ABC):
    """Class responsible for converting table operation objects to raw SQL strings
    in a specific dialect. Each of the supported database providers should inherit
    a converter from this base class and implement its methods"""

    @staticmethod
    @abstractmethod
    def convert_table_operation_to_SQL(operation: TableOperation) -> str:
        """Convert a given table operation object to raw SQL in a specific dialect"""


class ConstraintSQLConverter(ABC):
    """Class responsible for converting constraint objects to raw SQL strings
    in a specific dialect. Each of the supported database providers should inherit
    a converter from this base class and implement its methods"""

    @staticmethod
    @abstractmethod
    def convert_constraint_to_SQL(constraint: Constraint) -> str:
        """Convert a given constraint object to raw SQL in a specific dialect"""



class IrreversibleOperationError(TypeError):
    """Error thrown after an attempt to reverse an irreversible
    database operation"""


class UnsupportedOperationError(ValueError):
    """Error thrown after an attempt to convert or perform an operation
    unsupported by the current adapter or converter"""



import pytest
from skibidi_orm.migration_engine.data_mutator.sqlite3_data_mutatorr import (
    SQLite3DataMutator,
)
from skibidi_orm.migration_engine.db_config.sqlite3_config import SQLite3Config
from skibidi_orm.migration_engine.db_inspectors.sqlite3_inspector import (
    SQLite3Inspector,
)
from skibidi_orm.migration_engine.studio.utils.get_db_seeder import get_db_mutator


def test_get_db_seeder_sqlite3_error_when_config_not_exists():
    with pytest.raises(ReferenceError):
        get_db_mutator(db_inspector=SQLite3Inspector())


def test_get_db_seeder_sqlite3():
    SQLite3Config(db_path="some/path/to/db.db")
    seeder = get_db_mutator(db_inspector=SQLite3Inspector())
    assert isinstance(seeder, SQLite3DataMutator)



from skibidi_orm.migration_engine.adapters.base_adapter import (
    BaseTable,
    BaseColumn,
)
from typing import Any
from skibidi_orm.migration_engine.operations.column_operations import (
    ColumnOperation,
    AddColumnOperation,
    DeleteColumnOperation,
)

from skibidi_orm.migration_engine.operations.table_operations import (
    TableOperation,
    CreateTableOperation,
    DeleteTableOperation,
    RenameTableOperation,
)

from skibidi_orm.migration_engine.state_manager.i_state_manager import IStateManager


class StateManager[TTable: BaseTable[BaseColumn[Any]]](IStateManager):
    def __init__(
        self,
        db_tables: list[TTable],
        schema_tables: list[TTable],
    ) -> None:
        self.db_tables = db_tables
        self.schema_tables = schema_tables

        self.serviced_tables: list[TTable] = []

        self.operations: list[TableOperation | ColumnOperation] = []

        self.analyze_schemas()

    def analyze_schemas(self):
        self.get_delete_tables_operations()
        self.get_create_tables_operations()
        self.get_delete_columns_operations()
        self.get_create_columns_operations()

    def get_operations(self) -> list[TableOperation | ColumnOperation]:
        return self.operations

    def get_create_tables_operations(self) -> None:

        t_db_names = set([t.name for t in self.db_tables])

        for s_table in self.schema_tables:
            if s_table.name not in t_db_names:
                if s_table not in self.serviced_tables:
                    self.operations.append(CreateTableOperation(s_table))
                    self.serviced_tables.append(s_table)

    def get_delete_tables_operations(self) -> None:

        class TableFound(Exception):
            pass

        t_schema_names = set([t.name for t in self.schema_tables])
        for db_table in self.db_tables:
            try:
                if db_table.name not in t_schema_names:
                    for s_table in self.schema_tables:
                        s_column_names = set([c.name for c in s_table.columns])
                        db_column_names = set([c.name for c in db_table.columns])

                        if s_column_names == db_column_names:
                            self.operations.append(
                                RenameTableOperation(db_table, s_table.name)
                            )
                            self.serviced_tables.append(s_table)
                            raise TableFound
                    if db_table not in self.serviced_tables:
                        self.operations.append(DeleteTableOperation(db_table))
                        self.serviced_tables.append(db_table)
            except TableFound:
                pass

    def get_create_columns_operations(self):

        db_table_dict = {t.name: t for t in self.db_tables}
        schema_table_dict = {t.name: t for t in self.schema_tables}

        for s_table_name, s_table in schema_table_dict.items():
            if s_table not in self.serviced_tables:
                if s_table_name in db_table_dict.keys():

                    table_db_columns = db_table_dict[s_table_name].columns
                    table_schema_columns = s_table.columns

                    for s_column in table_schema_columns:
                        if s_column not in table_db_columns:
                            self.operations.append(
                                AddColumnOperation(
                                    db_table_dict[s_table_name], s_column
                                )
                            )

    def get_delete_columns_operations(self):

        db_table_dict = {t.name: t for t in self.db_tables}
        schema_table_dict = {t.name: t for t in self.schema_tables}

        for db_table_name, db_table in db_table_dict.items():
            if db_table not in self.serviced_tables:
                if db_table_name in schema_table_dict.keys():

                    table_db_columns = db_table.columns
                    table_schema_columns = schema_table_dict[db_table_name].columns

                    for db_column in table_db_columns:
                        if db_column not in table_schema_columns:
                            self.operations.append(
                                DeleteColumnOperation(db_table, db_column)
                            )

    # def get_relations_operations(self):
    #     # TODO: implement using foreign keys
    #     for relation in self.schema_relations:
    #         if relation not in self.db_relations:
    #             pass  # add crea
